-- Compute LL(1) predict sets.
-- (c) 2010 Bjoern B. Brandenburg.
-- All rights reserved.
--
-- This solution is based on slides 90-96 in Lecture 5 of
-- the COMP 524 Spring 2010 slides.

-- import a set abstract data type
import qualified Data.Set as S
-- import a map abstract data type
import qualified Data.Map as M
-- import Maybe for catMaybe
import Data.Maybe
-- import Control.Monad for forM_
import Control.Monad

-- Simple type synonyms for terminals
-- and non-terminal symbol names.
type Token = String
type Name  = String

-- A symbol is either a terminal token
-- or a non-terminal.
data Symbol = Terminal    Token
            | NonTerminal Name
              deriving (Show, Ord, Eq)

-- A production defines a head in terms of
-- a sequence of symbols. An empty list encodes
-- epsilon.
data Production = Production Name [Symbol]
                  deriving (Eq, Ord)

-- A Grammar is simply a list of productions.
type Grammar = [Production]

-- This type represents a set that we are trying
-- to compute. For example,
--   First (NonTerminal "X")
-- represents the set FIRST(X).
data Subset = First   Symbol
            | Follow  Symbol
              deriving (Show, Ord, Eq)


-- Initial set definitions.
-- Given the name of a set (a Subset value), return an
-- actual set with the initial members.
-- Only first sets for terminals are initially known.
initial :: Subset -> S.Set Token
initial (First (Terminal x)) = S.singleton x
initial _                    = S.empty

-- slide 93: inductive definition of FIRST(A)
-- Given a non-terminal name, check whether the given
-- production applies, and if so return a subset definition.
first :: Name -> Production -> Maybe Subset
first a (Production id def) =
-- Check if the production is for a, and if so return FIRST(x).
    if id == a
    then if null def
         -- yes, it's for a
         then Just (Follow $ NonTerminal a)  -- epsilon produciton
         else Just (First  $ head def)       -- not epsilon
    else Nothing

-- slide 95: inductive definition of FOLLOW(A)
-- Given a non-terminal name, check whether the given
-- production contains a reference to the name,
-- and if so return a subset definition.
follow :: Name -> Production -> Maybe Subset
follow a (Production pHead syms) = follow' syms
    where
      -- case 1: symbol name does not occur in this production
      follow' [] = Nothing
      -- case 2: symbol name occurs
      follow' ((NonTerminal id):xs) | id == a =
         Just $ if null xs
                -- case 2a: it's the last symbol -> use Follow
                then Follow $ NonTerminal pHead
                -- case 2b: it's not the last symbol -> use First
                else First (head xs)
      -- doesn't match, keep looking
      follow' (_:xs)  = follow' xs

-- Which subset of terminals is the actual predict set?
-- slide 92 of lecture 5
predict :: Production -> Subset
-- Case 1: epsilon production.
predict (Production head [])   = Follow (NonTerminal head)
-- Case 2: non-epsilon production, use the first symbol.
predict (Production _ (sym:_)) = First  sym

-- Define a subset in terms of included subsets.
define :: Subset -> Grammar -> [Subset]
-- slide 93: First of a terminal is just itself.
define t@(First  (Terminal    _)) = const [t]
-- slide 93: all sets included in a first set
define (First  (NonTerminal a)) = catMaybes . map (first  a)
-- slide 95: all sets included in a follow set
define (Follow (NonTerminal a)) = catMaybes . map (follow a)

-- Two type synonyms to make the the signatures a bit more
-- readable.
-- SetVals is a mapping from Subset names to actual sets
-- of tokens. This represents the "currently known" set definitions,
-- which may be incomplete.
-- SetDefs is a mapping from Subset names to Subset lists.
-- This represents the abstract definition of the subsets, i.e.,
-- the list of subsets represents the included subsets.
type SetVals = M.Map Subset (S.Set Token)
type SetDefs = M.Map Subset [Subset]

-- Given currently-known set definitions, and a list of subsets names,
-- create a new set that is the union of the given subset names.
-- Additionally, keep track of subsets for which definitions are
-- missing.
populate :: SetVals -> [Subset] -> [Subset] -> ([Subset], S.Set Token)
populate vals missing subsets =
    let
        lookup = flip M.lookup
        sets   = map (lookup vals) subsets
        missed = map snd $ filter (isNothing . fst) $ zip sets subsets
    in
      (missed ++ missing, (S.unions . catMaybes) sets)

-- Given a grammar and a tuple of abstract and concrete set definitions,
-- create new abstract and concrete set definitions. This is the
-- iteration step in which subset member are propagated.
propagate :: Grammar
          -> (SetDefs, SetVals)
          -> (SetDefs, SetVals)
propagate g (defs, vals) =
    let
        (missing, newVals) = M.mapAccum (populate vals) [] defs
        missDefs = M.fromList $ zip missing $ map (flip define g)  missing
        missVals = M.fromList $ zip missing $ map initial missing
    in
      (defs `M.union` missDefs, newVals `M.union` missVals)

-- Generic fixpoint search. Given a function and an initial
-- point, keep applying the function to its result until
-- the point doesn't change anymore.
fixpoint :: (Eq a) => (a -> a) -> a -> a
fixpoint f x0 =
    let
        pts = iterate f x0
        fix = dropWhile (uncurry (/=)) $ zip pts (tail pts)
    in
      (fst . head) fix

-- compute LL(1) predict sets for each production
predictSets :: Grammar -> [(Production, [Token])]
predictSets g =
    let
        -- first, figure out for each production what we need.
        sets      = map predict g
        -- Get initial actual set definitions. Most of these will
        -- simply be empty.
        iniVals   = M.fromList $ zip sets $ map initial sets
        -- Get initial abstract set definitions.
        -- They will probably be incomplete, but the iteration step
        -- will discover any missing subset definitions.
        iniDefs   = M.fromList $ zip sets $ map (flip define g) sets
        -- Compute predict sets. Correct predict sets are given
        -- by the fixpoint of the propagate function: if propagation
        -- doesn't change the sets anymore, then we have found
        -- consistent predict sets.
        (_, vals) = fixpoint (propagate g) (iniDefs, iniVals)
    in
      -- extract the found predict sets
      zip g $ map (S.toList . fromJust . (flip M.lookup vals)) sets


-- Non-LL(1) test grammar.
-- This grammar is left-recursive, and thus not LL(1).
-- This is revealed by the non-disjoint predict sets that are being
-- computed.
testGrammar = [ Production "Start" [NonTerminal "Expr", Terminal "EOF" ]
              , Production "Expr" [ NonTerminal "Expr", Terminal "+" ]
              , Production "Expr" [ NonTerminal "Term" ]
              , Production "Term" [ NonTerminal "Term", Terminal "*"
                                  , NonTerminal "Primary"]
              , Production "Term" [ NonTerminal "Primary" ]
              , Production "Primary" [ NonTerminal "Funcall" ]
              , Production "Primary" [ Terminal "Id"]
              , Production "Primary" [ Terminal "("
                                     , NonTerminal "Expr"
                                     , Terminal ")" ]
              , Production "Funcall" [ Terminal "Id"
                                     , Terminal "("
                                     , NonTerminal "OptArgs"
                                     , Terminal ")" ]
              , Production "OptArgs" []
              , Production "OptArgs" [ NonTerminal "Args" ]
              , Production "Args"    [ NonTerminal "Expr" ]
              , Production "Args"    [ NonTerminal "Args"
                                     , Terminal ","
                                     , NonTerminal "Expr" ]
              ]

-- A LL(1) test grammar.
testGrammar2 = [ Production "Start" [ NonTerminal "E", Terminal "EOF" ]
               , Production "E" [ NonTerminal "T"
                                , NonTerminal "E'" ]
               , Production "E'" [ Terminal "+"
                                 , NonTerminal "T"
                                 , NonTerminal "E'"]
               , Production "E'" []
               , Production "T" [ NonTerminal "F"
                                , NonTerminal "T'" ]
               , Production "T'" [ Terminal "*"
                                 , NonTerminal "F"
                                 , NonTerminal "T'"]
               , Production "T'" []
               , Production "F" [ Terminal "("
                                 , NonTerminal "E"
                                 , Terminal ")"]
               , Production "F" [ Terminal "id" ]
               ]

-- pretty printing productions
showProd :: Production -> String
showProd (Production id [])  = id ++ " -> <eps>"
showProd (Production id def) = id ++ " ->" ++ foldl join "" def
    where join s (Terminal x)    = s ++ " '" ++ x ++ "'"
          join s (NonTerminal x) = s ++ " " ++  x

instance Show Production where
    show = showProd

-- pad strings with spaces to make them look centered
centered :: Int -> String -> String
centered width s =
    let
        n    = length s
        pre  = (width - n) `div` 2
        post = pre + (width - n) `mod` 2
    in
      (replicate pre ' ') ++ s ++ (replicate post ' ')

-- Print one production and its corresponding predict set.
printProd :: Int -> Production -> [Token] -> IO ()
printProd width p tok = putStrLn $ "PREDICT(" ++ pstr ++ ") = { " ++
                        (s tok) ++ " }"
    where s []  = ""
          s lst = foldl1 (\s1 s2 -> s1 ++ ", " ++ s2) $ map quote lst
          pstr = centered (width + 2) (show p)
          quote str = "'" ++ str ++ "'"

-- Print all productions.
printPredictSets :: Grammar -> IO ()
printPredictSets g = do
  let width = maximum (map (length . show) g)
  forM_ (predictSets g) (uncurry $ printProd width)
